"""
==============
Scintillations
==============

Atmospheric turbulence causes fluctuations in the sound speed which in effects causes fluctuations in the amplitude and phase of the sound pressure.
This module contains functions for generating and analyzing sequences of scintillations that follow a Gaussian spectrum,
"""

import scintillations.common
import scintillations.sequence
import scintillations.stream
